//
//  PseudokoDocument.m
//  Pseudoko
//
//  Created by David Phillip Oster on 1/8/07.
//  Copyright David Phillip Oster 2007 . 
//

#import "PseudokoDocument.h"
#import "PuzzleProxy.h"
#import "SolvableStatusTransformer.h"

static SolvableStatusTransformer *gSolvableStatusTransformer = nil;

@implementation PseudokoDocument
+ (void)initialize {
  gSolvableStatusTransformer = [[SolvableStatusTransformer alloc] init];
  [[gSolvableStatusTransformer class] setValueTransformer:gSolvableStatusTransformer forName:@"SolvableStatusTransformer"];
}


- (id)init{
  if (nil != (self = [super init])) {
    moves = [[NSMutableArray alloc] init];

    NSMutableArray *array = [[NSMutableArray alloc] init];
    for(int  i = 0; i < 9*9; ++i) {
      [array addObject:[[PuzzleProxy alloc] initWithIndex:i document:self]];
    }
    puzzleProxys = array;
  }
  return self;
}

// Note 1:
// I can get a consistent crash by creating a new document, typing a "1" in
// the first cell, clicking on another cell, typing a "1", clicking on third cell,  
// then an Undo,  Close, Don't Save.
// The crash was a BAD ACCESS in -[NSMatrix setNeedsDisplayInRect:], after the
// window was hidden and dealloc was called and was 100% repeatable.
// See Bugs.txt in this project's folder for a stack trace.
// http: comment below documents the cure.
- (void)dealloc {
  NSUndoManager *undo = [self undoManager];
  [undo removeAllActions];
  NSNotificationCenter *nc = [NSNotificationCenter defaultCenter];
  [nc removeObserver:self];

  // undo what we programmatically did in windowControllerDidLoadNib:
  for(int k = 0, j = 0; j < 9; ++j) {
    for(int i = 0; i < 9; ++i) {
      id cell = [puzzleMatrix cellAtRow:j column:i];
      PuzzleProxy *proxy = [puzzleProxys objectAtIndex:k];
      [cell unbind:NSValueBinding];
      [cell unbind:NSTextColorBinding];
      [cell unbind:NSEditableBinding];
      [proxy removeObserver:self forKeyPath:@"value"];
      [proxy removeObserver:self forKeyPath:@"isEditable"];
      ++k;
    }
  }
  // http://www.cocoabuilder.com/archive/message/cocoa/2006/7/9/167168
  [puzzleMatrix abortEditing];
  [puzzleProxys release];
  [moves release];
  [super dealloc];
}

// dump the puzzle for debugging
- (NSString *)description {
  NSMutableString *s = [NSMutableString string];
  for (int row = 0; row < 9; ++row) {
    for (int column = 0; column < 9; ++column) {
      [s appendFormat:@"%2d%@", (0xF & puzzle[row][column]), 8 != column ? @" " : @"\n"];
    }
  }
  return s;
}

- (NSString *)windowNibName{
  return @"PseudokoDocument";
}

- (void)setIntValue:(int)val row:(int)row column:(int)column {
  puzzle[row][column] = (val | (puzzle[row][column] & ~kValueMask));
}

- (void)changeKeyPath:(NSString *)keyPath ofObject:(id)obj toValue:(id)newValue {
  [obj setValue:newValue forKeyPath:keyPath];
}

- (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context {
  NSUndoManager *undo = [self undoManager];
  id oldValue = [change valueForKey:NSKeyValueChangeOldKey];
  [[undo prepareWithInvocationTarget:self] changeKeyPath:keyPath ofObject:object toValue:oldValue];
  if ([keyPath isEqual:@"value"] && ! ([undo isUndoing] || [undo isRedoing])) {
    [undo setActionName:NSLocalizedString(@"Move", @"")];
  }
}


- (void)windowControllerDidLoadNib:(NSWindowController *) aController{
  [super windowControllerDidLoadNib:aController];
  for(int k = 0, j = 0; j < 9; ++j) {
    for(int i = 0; i < 9; ++i) {
      id cell = [puzzleMatrix cellAtRow:j column:i];
      PuzzleProxy *proxy = [puzzleProxys objectAtIndex:k];
      [cell bind:NSValueBinding toObject:proxy withKeyPath:@"value" options:nil];
      [cell bind:NSTextColorBinding toObject:proxy withKeyPath:@"color" options:nil];
      [cell bind:NSEditableBinding toObject:proxy withKeyPath:@"isEditable" options:nil];
      [proxy addObserver:self forKeyPath:@"value" options:NSKeyValueObservingOptionOld context:nil];
      [proxy addObserver:self forKeyPath:@"isEditable" options:NSKeyValueObservingOptionOld context:nil];
      ++k;
    }
  }
}

// ** Actions (sent from menu bar to first responder, so I.B. doesn't need to know about them here)
- (BOOL)validateMenuItem:(NSMenuItem *)anItem {
  SEL sel = [anItem action];
  if (sel == @selector(toggleCreatorMode:)) {
    [anItem setState:creatorMode];
  } else if (sel == @selector(lock:) || sel == @selector(unlock:)) {
    return creatorMode;
  }
  return [super validateMenuItem:anItem];
}

- (void)toggleCreatorMode:(id)sender {
  creatorMode = ! creatorMode;
}

// Finalize any in-progress editing, and if that created an implicit undogroup, close the
// group so the next command, if any, will be in its own group.
// But, if finalizing put up an error sheet, then return NO, so caller will know to abort the command.
- (BOOL)endEditing {
  [[puzzleMatrix window] endEditingFor:nil]; //user types '3<Solv>' he means the three to be considered.
  BOOL canContinue = (nil == [[puzzleMatrix window] attachedSheet]);
  if (canContinue) {
    NSUndoManager *undo = [self undoManager];
    if ( ! ([undo isUndoing] || [undo isRedoing]) && 1 == [undo groupingLevel]) {
      [undo endUndoGrouping];
      [undo beginUndoGrouping];
    }
  }
  return canContinue;
}


- (void)lock:(id)sender {
  if ([self endEditing]) { //user types '3<lock>' he means the three to be locked.
    for(int  i = 0; i < 9*9; ++i) {
      PuzzleProxy *proxy = [puzzleProxys objectAtIndex:i];
      if (0 != [proxy value]){
        [proxy setIsEditable:NO];
      }
    }
    [[self undoManager] setActionName:NSLocalizedString(@"Lock", @"")];
  }
}

- (void)unlock:(id)sender {
  if ([self endEditing]) { //user types '3<lock>' he means the three to be locked.
    for(int  i = 0; i < 9*9; ++i) {
      PuzzleProxy *proxy = [puzzleProxys objectAtIndex:i];
      [proxy setIsEditable:YES];
    }
    [[self undoManager] setActionName:NSLocalizedString(@"Unlock", @"")];
  }
}

- (BOOL)isProvenNotSolvable {
  return isProvenNotSolvable;
}

- (void)setIsProvenNotSolvable:(BOOL)solvable {
  isProvenNotSolvable = solvable;
}



// ** Overrides for i/o

- (NSData *)dataOfType:(NSString *)typeName error:(NSError **)outError{
  return [NSData dataWithBytes:puzzle length:sizeof puzzle];
}

- (BOOL)readFromData:(NSData *)data ofType:(NSString *)typeName error:(NSError **)outError{
  [data getBytes:puzzle length:sizeof puzzle];
  return YES;
}

@end
/*
Permission is hereby granted, free of charge, to any person obtaining
a copy of these files and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
